# -*- Gap -*- 

BindGlobal("PU_DEFAULT_TIMEOUT",60);
BindGlobal("PU_LONG_TIMEOUT",3600);

DeclareInfoClass("InfoProcess");

InstallMethod(ReadLineWithTimeout,[IsInputStream,IsInt],
        function(p,timeout)
    local fd,L,ready,line,partial;
    

    fd:=FileDescriptorOfStream(p);
    L:=[fd];
    line:="";
    while(Chomp(line)=line) do
        ready:=UNIXSelect(L,[],[],timeout,0);
        Info(InfoProcess,3,"select = ",ready);
        if (ready<0) then
            Info(InfoProcess,1,"Error in select");
            return fail;
        fi;
        if (ready=0) then
            Info(InfoProcess,1,"Read from server timed out");
            return fail;
        fi;
        partial:=ReadLine(p);
        Info(InfoProcess,3,"partial = ",partial);
        line:=Concatenation(line,partial);
    od;
    Info(InfoProcess,3,"Read ",line);
    return line;
end);

InstallOtherMethod(ReadLineWithTimeout,[IsInputStream],
        function(p)
    return ReadLineWithTimeout(p,PU_LONG_TIMEOUT);
end);

InstallMethod(Expect,[IsInputStream,IsString,IsInt],
        function(p,s,timeout)
    local line;
    line:=ReadLineWithTimeout(p,timeout);
    line:=Chomp(line);
    return (line=s);
end);

InstallOtherMethod(Expect,[IsInputStream,IsString],
        function(p,s)
    return Expect(p,s,PU_DEFAULT_TIMEOUT);
end);

InstallMethod(Insist,[IsInputStream,IsString,IsInt],
        function(p,s,t)
    if (not Expect(p,s,t)) then
       Error("expected ", s , "from process ", p);
    fi;
end);
  

InstallOtherMethod(Insist,[IsInputStream,IsString],
        function(p,s)
    Insist(p,s,PU_DEFAULT_TIMEOUT);
end);


InstallMethod(SendList,[IsOutputStream,IsList],  
        function(p,l)
    local line;
    line:=JoinStringsWithSeparator(List(l,String)," ");
    WriteAll(p,Concatenation(" ",line));
end);

InstallGlobalFunction(SendLine,
        function(arg)
    SendList(arg[1],arg{[2..Size(arg)]});
    WriteLine(arg[1],"");
end);

InstallGlobalFunction(SendItems,
        function(arg)
    SendList(arg[1],arg{[2..Size(arg)]});
end);


InstallMethod(ReadList,[IsInputStream,IsPosInt],
        function(p,timeout)
    local line;
    line:=ReadLineWithTimeout(p,timeout);
    if (line=fail) then 
        return fail;
    fi;
    Chomp(line);
    return SplitString(line," ");
end);

InstallOtherMethod(ReadList,[IsInputStream],function(p)
    return ReadList(p,PU_LONG_TIMEOUT);
end);

InstallMethod(StringToRatVec,[IsString],function(string)
    local list;
    string:=Chomp(string);
    list:=SplitString(string," ");
    if (Int(list[1]) =fail) then
        Error("vector string must start with integer length");
    fi;

    list:=List(list{[2..Size(list)]},Rat);
    Info(InfoProcess,3,"String to rat vec returning ",list);
    return list;
end);