
# Some not too clever ideas for finding canonical representatives.
# Nothing here is tied to polyhedra, etc...
RequirePackage("require_file");
RequireFile("gram.gap");

InfoCanon:=NewInfoClass("Canon");
SetInfoLevel(InfoCanon,1);

ExhaustiveCanon:=function(Grp,element,action,compare)
    local r,O,canon;
    O:=Orbit(Grp,element,action);
    
    canon:=element;
    for r in O do
        if compare(r,canon) then
            canon:=r;
        fi;
    od;
    return canon;
end;

StabOrbitMinimum:=function(Grp,element,compare)
    local r,O,stab,canon,steps;
    stab:=Stabilizer(Grp,[1..Maximum(element)],OnSets);
    Info(InfoCanon,3,"Stabilizer order = ",Order(stab),"\n");
    O:=Orbit(stab,element,OnSets);
    canon:=O[1];
    for r in O do
        if compare(r,canon) then
            canon:=r;
        fi;
    od;
    return canon;
end;




GreedyDescent:=function(G,element,action,compare,steps)

    local smallest,lastIndex,i,j,K,stuck,pi,generators;
    
    generators:=GeneratorsOfGroup(G);
    
    smallest:=element;
    K:=smallest;
    for i in [1..steps] do
        Info(InfoCanon,4,"Greedy descent attempt ",i);
        pi:=PseudoRandom(G);
        # first jump. It shouldn't matter where we start?
        K:=action(element,pi);
        if compare(K,smallest) then
            smallest:=K;
        fi;
        repeat
            stuck:=true;
            for j in [1..Size(generators)] do
                K:=action(smallest,generators[j]);
                if compare(K,smallest) then
                    smallest:=K;
                    stuck:=false;
                    Info(InfoCanon,4,"new rep ",smallest);
                fi;
            od;
            if (stuck and Maximum(smallest)<LargestMovedPoint(G)) then
                K:=StabOrbitMinimum(G,smallest,compare);
                
                if compare(K,smallest) then
                    smallest:=K;
                    stuck:=false;
                    Info(InfoCanon,4,"new rep ",smallest);
                fi;
            fi;
        until (stuck);
    od;
    return smallest;
end;

StabCanon:=function(Grp,element,minGround,maxGround,knownCanons)
    local O,r,stab,heuristic,canon,ground,leftOut,j;
    
    ground:=element;
    leftOut:=Difference(MovedPoints(Grp),ground);
    
    Info(InfoCanon,3,"known Canons= ",Size(knownCanons),"\n");
    while(Size(ground)< minGround) do
        j:=PseudoRandom(leftOut);
        Info(InfoCanon,4,"preprocess: adding j = ",j);
        ground:=Union(ground,[j]);
        RemoveSet(leftOut,j);
    od;
    
    repeat
        stab:=Stabilizer(Grp, ground,OnSets);
        Info(InfoCanon,3,"Stabilizer Size =",Order(stab));
        Info(InfoCanon,4,"Ground =",ground);
        O:=Orbit(stab,element,OnSets);
        canon:=RepresentativeSmallest(O);
        if (canon in knownCanons) then
            Info(InfoCanon,2,"Found old canon in orbit of size ",Size(O), 
                 " ground size=", Size(ground) );
            return canon;
        fi;
        if (Size(leftOut)>0) then
            j:=PseudoRandom(leftOut);
            Info(InfoCanon,4,"j = ",j);
            ground:=Union(ground,[j]);
            RemoveSet(leftOut,j);
        fi;
    until  Size(ground)>=maxGround;
    Info(InfoCanon,2,"Found new canon in orbit of size ",Size(O), 
                 " ground size=", Size(ground) );
    return canon;
end;


GreedyCanon:=function(element,generators,action)
    local smallest,lastIndex,i,K,stuck;
    
    smallest:=element;
    lastIndex:=-1;
    repeat
        stuck:=true;
        for i in [1..Size(generators)] do
            K:=action(smallest,generators[i]);
            if (K < smallest) then
                smallest:=K;
                lastIndex:=i;
                stuck:=false;
            fi;
        od;
    until (stuck);
    return smallest;
end;


GreedyRandomCanon:=function(G,element,action,steps)

    local smallest,lastIndex,i,K,stuck,pi,generators;
    
    generators:=GeneratorsOfGroup(G);
    
    smallest:=element;
    K:=smallest;
    for i in [1..steps] do
        pi:=PseudoRandom(G);
        # first jump. It shouldn't matter where we start?
        K:=action(element,pi);
        if (K < smallest) then
            smallest:=K;
        fi;
        repeat
            stuck:=true;
            for i in [1..Size(generators)] do
                K:=action(smallest,generators[i]);
                if (K < smallest) then
                    smallest:=K;
                    lastIndex:=i;
                    stuck:=false;
                fi;
            od;
        until (stuck);
    od;
    return smallest;
end;

orbitReps:=function(group,set,action,size)
    local reps1,reps2,orbs,e1,e2,work;
    if (Size(set)>size) then
        Info(InfoCanon,3,"splitting ",Size(set),"\n");
        reps1:=orbitReps(group,set{[1..QuoInt(Size(set),2)]},action,size);
        reps2:=orbitReps(group,set{[QuoInt(Size(set),2)+1..Size(set)]},action,size);
        Info(InfoCanon,3,"merging ", Size(reps1),"+",Size(reps2),"\n");
        work:=Set(reps2);
        for e1 in reps1 do
            for e2 in work do
                if (e1 <> e2) and RepresentativeAction(group,e1,e2,action) <> fail then
                    RemoveSet(work,e2);
                fi;
            od;
        od;
        return Union(reps1,work);
    else
        Info(InfoCanon,3,"quadratic scanning\n");
        work:=Set(set);
          for e1 in work do
            for e2 in work do
                if (e1 <> e2) and RepresentativeAction(group,e1,e2,action) <> fail then
                    RemoveSet(work,e2);
                fi;
            od;
        od;
        return work;
    fi;
end;


gramSift:=function(group,bases,IPM,action)
    local dict,basis,vec,list,output,i,merged;
    
    dict:=newDictionary();
    
    for basis in bases do
        vec:=fastGramVec(basis,IPM);
        addDictionary(dict,vec,basis);
    od;
    output:=[];
    
    list:=dictionaryValues(dict);
    for i in [1..Size(list)] do
        Info(InfoCanon,2,"processing list ",i, " of size ", Size(list[i]));
        merged:=orbitReps(group,list[i],action,4);
        Info(InfoCanon,2,"unique representatives= ", Size(merged));
        output:=Union(output,merged);
    od;
    return output;

end;