
RequirePackage("require_file");
AddToLoadPath("~/research/gap-lrslib/src");
RequireFile("matrix-utils.gap");

InfoRatio:=NewInfoClass("Ratio");
                  
slackSys:=function(V)
    local A0,bbar,Ab0,A1,b,b0,A2, A3,A,Vt,Vp,d,m,pi,initial_cobasis;
    
    d:=Size(V[1])-1;
    m:=Size(V);
    
    
    # add slacks, 

    A0:=adjoin(columns(V,[2..(d+1)]), -1*IdentityMat(m));
    b0:=-1*column(V,1);
    return rec(A:=A0,b0:=b0,A0:=A0,b:=b0,d:=d,m:=m);
end;

lslackSys:=function(V)
    local A0,bbar,Ab0,A1,b,b0,A2, A3,A,Vt,Vp,d,m,pi,initial_cobasis;
    
    d:=Size(V[1])-1;
    m:=Size(V);
    
    # reorder rows so that the last d define a vertex
    if (RankMat(V{[m-d+1..m]})=d) then
        initial_cobasis:=[m-d+1..m];
    else
        initial_cobasis:=LinearIndependentColumns(TransposedMat(columns(V,[2..d+1])));
    fi;
    Info(InfoRatio,2,"initial cobasis =",initial_cobasis);
    
    pi:=MappingPermListList(initial_cobasis,[m-d+1..m]);
    Vp:=Permuted(V,pi);
    
    # add slacks, 

    A0:=adjoin(columns(Vp,[2..(d+1)]), -1*IdentityMat(m));
    
    bbar:=column(Vp,1);
    
    # solve for the first m decision+slacks    
    Ab0:=columns(A0,[1..m]);
    A1:=(1/Ab0)*A0;
    
    #augment with objective row and column
    
    b:=Concatenation([[0],-1/Ab0*bbar]);

    A2:=adjoin(TransposedMat([constvec(0,m)]), A1);
    A3:=stack([Concatenation([1],constvec(0,m), constvec(1,d))],A2);

    A:=A3;
    return rec(T:=Ab0,A0:=A0,A:=A,b:=b,b0:=-1*column(V,1),d:=d,m:=m,pi:=pi, initial_cobasis:=[m-d+1..m]);
end;



candidates:=function(s,dict)
    # leave the decision variables in the basis
    return Filtered([dict.d+2..dict.m+1],j->dict.Ap[j][s]>0);
end;


ratios:=function(s,rows,beta,Ap)
    local ratios,min,indices;
#entering variable, as a column of A_B^{-1}*A_N

    ratios:=List(rows,i->beta[i]/Ap[i][s]);
    
    min:=Minimum(ratios);
    Info(InfoRatio,3,"minimum ratio ",min, " from ", ratios,"\n");
    indices:=Filtered([1..Size(rows)], i->ratios[i]=min);
    return rows{indices};
end;

makedict:=function(cobasis,sys)
    
    local N,B,Ab,Abi,An,beta,Ap,m,d,Rows,i;
    
    m:=sys.m;
    d:=sys.d;
    
    N:=List(cobasis,j->j+d);
    
    B:=Difference([1..m+d],N);
    Info(InfoRatio,3,"B=",B);
    
    Ab:=columns(sys.A0,B);
    if (RankMat(Ab)<Size(Ab)) then
        Error("rank Ab=",RankMat(Ab));
    fi;
    Abi:=1/Ab;
    
    An:=columns(sys.A0,N);
    
    beta:=(Abi)*sys.b0;

    Ap:=(1/Ab)*An;
    
    return rec(
               style:="raw",
               N:=N,
               B:=B,Abi:=Abi,
               An:=An,
               beta:=beta,
               Ap:=Ap,
               d:=sys.d,m:=sys.m);
end;

# do brute force orbit ratio tests without fancy compression of 
# basis inverse. Note that this version of the routine depends on 
# _raw_ dictionaries, which in turn depend on _raw_ slack system.

orbitratio:=function(s,dict,orbits,signs)
    local C,rows,j,beta,orb;
    
    C:=Filtered([dict.d+1..dict.m],j->dict.Ap[j][s]>0);
    Info(InfoRatio,3,"orbit candidate rows=",C,"\n");    
    rows:=ratios(s,C,dict.beta,dict.Ap);
    Info(InfoRatio,3,"orbit initial rows=",rows,"\n");    
    j:=1;
    while (Size(rows)>1 and (j<=Size(orbits))) do
        orb:=orbits[j];
        Info(InfoRatio,3,"checking orbit[",j,"]=", orb,"\n");
        # We have to negate the sign because...????
        beta:=-1*signs[j]*sumcols(dict.Abi,orb);
        rows:=ratios(s,rows,beta,dict.Ap);
        Info(InfoRatio,3,"rows = ",rows,"\n");
        j:=j+1;
    od;

    return List(rows,i->dict.B[i]-dict.d);
end;


slowOrbitRatio:=function(col,cobasis,sys,orbits,signs)
    local dict;
    dict:=makedict(cobasis,sys);
    return orbitratio(col,dict,orbits,signs);
end;



lrsdict:=function(cobasis,sys)
    local N,B,Ab,Abi,An,beta,Ap,m,d,Rows,i;
    
    m:=sys.m;
    d:=sys.d;
    
    N:=List(cobasis,j->j+d+1);
    
    B:=Difference([1..m+d+1],N);

    Ab:=columns(sys.A,B);
    
    Abi:=1/Ab;
    
    An:=columns(sys.A,N);
    
    beta:=(Abi)*sys.b;

    Ap:=(1/Ab)*An;
    
    Rows:=[];
    for i in [1..Size(B)] do
        Rows[B[i]]:=i;
    od;
    
    return rec(
               style:="lrs",
               Rows:=Rows,
               T:=sys.T,
               N:=N,
               B:=B,Abi:=Abi,
               An:=An,
               pi:=sys.pi,
               beta:=beta,
               Ap:=Ap,
               d:=sys.d,m:=sys.m);
end;

###################################################################
# experimental stuff below here

basisRank:=function(cobasis,sys)
    local N,B,Ab;
    N:=List(cobasis,j->j+sys.d);
    
    B:=Difference([1..sys.m+sys.d],N);
    Info(InfoRatio,3,"B=",B);
    
    Ab:=columns(sys.A0,B);
    return RankMat(Ab);
end;

basisRank2:=function(cobasis,sys)
    local N,B,Ab;
    N:=List(cobasis,j->j+sys.d+1);
    
    B:=Difference([1..sys.m+sys.d+1],N);
    Info(InfoRatio,3,"B=",B);
    
    Ab:=columns(sys.A,B);
    return RankMat(Ab);
end;

bruteratio:=function(s,dict)
    local C,rows,j;
    C:=candidates(s,dict);

    rows:=ratios(s,C,dict.beta,dict.Ap);
    Info(InfoRatio,3,"brute initial rows=",rows,"\n");    
    j:=1;
    while (Size(rows)>1) do
        Info(InfoRatio,3,"rows = ",rows,"\n");
        Info(InfoRatio,3,"checking inverse basis column ",j,"\n");
        rows:=ratios(s,rows,column(dict.Abi,j),dict.Ap);

        j:=j+1;
    od;
    return rows[1];
end;


# as per orbitratio, but attempt to deal with permuted rows, etc...

lorbitratio:=function(s,dict,orbits,signs)
    local C,rows,j,beta,Pi,xAbi,orb;
    
    
    if (dict.style <> "lrs") then
        Error("expecting lrs style dictionary");
    fi;
    
    
    xAbi:=dict.Abi*Permuted(dict.T,dict.pi);
    Info(InfoRatio,3,"Transformed basis inverse=",xAbi);
    C:=candidates(s,dict);
    Info(InfoRatio,3,"orbit candidate rows=",C,"\n");    
    rows:=ratios(s,C,dict.beta,dict.Ap);
    Info(InfoRatio,3,"orbit initial rows=",rows,"\n");    
    j:=1;
    while (Size(rows)>1 and (j<=Size(orbits))) do
        orb:=orbits[j];
        Info(InfoRatio,3,"checking orbit[",j,"]=", orb,"\n");
        # We have to negate the sign because we are implicitly changing 
        # 1 + Ax >= 0 to Ax - Iy = -1, I guess
        beta:=-1*signs[j]*sumcols(xAbi,orb);
        rows:=ratios(s,rows,beta,dict.Ap);
        Info(InfoRatio,3,"rows = ",rows,"\n");
        j:=j+1;
    od;
    return List(rows,i->OnPoints(dict.B[i]-1,dict.pi^(-1))-dict.d);
end;

lrsratio:=function(s,dict)
    local rows,j,cindex,C,bindex;
    
    C:=candidates(s,dict);
    rows:=ratios(s,C,dict.beta,dict.Ap);
    
    Info(InfoRatio,3,"lrs initial rows=",rows,"\n");

    cindex:=1;
    bindex:=dict.d+2;
    while (Size(rows)>1) do
        Info(InfoRatio,4,"cindex=",cindex," bindex=",bindex,"\n");
        if (dict.B[bindex]=bindex+cindex-1 ) then
            if (rows[1]=bindex) then
                Info(InfoRatio,3,"removing identity col ",bindex, "\n");
                Remove(rows,1);
            fi;
            bindex:=bindex+1;
        else
            Info(InfoRatio,3,"Checking non-identity column ",bindex," as cobasis column ",cindex,"\n");
            rows:=ratios(s,rows,column(dict.Ap,cindex),dict.Ap);
            cindex:=cindex+1;
        fi;
        Info(InfoRatio,3,"rows = ",rows,"\n");
    od;
    return rows[1];
end;

feasible:=function(dict)
    return ForAll(dict.beta{[dict.d+2..dict.m+1]},b->b>=0);
end;

rfeasible:=function(dict)
    return ForAll(dict.beta{[dict.d+1..dict.m]},b->b>=0);
end;

# return the _infeasible_ bases for system sys
checkBases:=function(bases,sys)
    return Filtered(bases,b->not feasible(makedict(b,sys)));
end;

compare:=function(bases,bindex,col)
    local cobasis,dict,lrows,brows;
    
    cobasis:=bases[bindex];

    dict:=makedict(cobasis);

    brows:=bruteratio(col,dict);
    Info(InfoRatio,3,"brute=",brows,"\n");
    lrows:=lrsratio(col,dict);
    Info(InfoRatio,3,"lrs=",lrows,"\n");
    return lrows=brows;
end;

lrsEntering:=function(lrs,cobasis)
    local leave,enter,entering;
    
    if (not LRSSetCobasis(lrs,cobasis)) then
        return fail;
    fi;

    entering:=[];
    for leave in cobasis do
        enter:=LRSLexRatio(lrs,leave);
        if (enter <> fail) then
            Add(entering,[enter]);
        fi;
    od;
    return entering;
end;

lrsAllEntering:=function(lrs,cobasis)
    local leave,enter,entering;
    
    if (not LRSSetCobasis(lrs,cobasis)) then
        return fail;
    fi;

    entering:=[];
    for leave in cobasis do
        enter:=LRSAllRatio(lrs,leave);
        if (enter <> fail) then
            Add(entering,enter);
        fi;
    od;
    return entering;
end;

# test the default case of orbitratio, which should be the 
# same as lexratio

lrsOrbitEntering:=function(lrs,cobasis)
    local leave,enter,entering;
    
    if (not LRSSetCobasis(lrs,cobasis)) then
        return fail;
    fi;

    entering:=[];
    for leave in cobasis do
        enter:=LRSOrbitRatio(lrs,leave);
        if (enter <> fail) then
            Add(entering,enter);
        fi;
    od;
    return entering;
end;

# note that thanks to my being too clever for my own good, lrsserv returns
# facet indices but bruteratio and lrsratio do not.
bruteEntering:=function(cobasis,dict)
    local enter,entering,col;
    
    if (not feasible(dict)) then
        return fail;
    fi;

    entering:=[];
    for col in [1..Size(cobasis)] do
        enter:=bruteratio(col,dict);
        if (enter <> fail) then
            Add(entering,enter);
        fi;
    od;
    return List(entering,j->dict.B[j]-dict.d-1);
end;

orbitEntering:=function(dict,O,sigma)
    local enter,entering,col;
    
    if (not rfeasible(dict)) then
        return fail;
    fi;

    entering:=[];
    for col in [1..dict.d] do
        enter:=orbitratio(col,dict,O,sigma);
        if (enter <> fail) then
            Add(entering,enter);
        fi;
    od;
    return entering;
end;

lorbitEntering:=function(dict,O,sigma)
    local enter,entering,col;
    
    if (not feasible(dict)) then
        return fail;
    fi;

    entering:=[];
    for col in [1..dict.d] do
        enter:=lorbitratio(col,dict,O,sigma);
        if (enter <> fail) then
            Add(entering,enter);
        fi;
    od;
    return entering;
end;

pseudoLrsEntering:=function(cobasis,dict)
    local enter,entering,col;
    
    if (not feasible(dict)) then
        return fail;
    fi;

    entering:=[];
    for col in [1..Size(cobasis)] do
        enter:=lrsratio(col,dict);
        if (enter <> fail) then
            Add(entering,enter);
        fi;
    od;
    return List(entering,j->dict.B[j]-dict.d-1);
end;

